"""
See openff/toolkit/utils/utils.py
"""

import functools
from collections.abc import Iterable

from openff.units import Quantity, unit


def string_to_quantity(quantity_string) -> str | int | float | Quantity:
    """Attempt to parse a string into a unit.Quantity.

    Note that dimensionless floats and ints are returns as floats or ints, not Quantity objects.
    """

    from tokenize import TokenError

    from pint import UndefinedUnitError

    try:
        quantity = Quantity(quantity_string)
    except (TokenError, UndefinedUnitError):
        return quantity_string

    # TODO: Should intentionally unitless array-likes be Quantity objects
    #       or their raw representation?
    if (quantity.units == unit.dimensionless) and isinstance(quantity.m, int | float):
        return quantity.m
    else:
        return quantity


def convert_all_strings_to_quantity(
    smirnoff_data: dict,
    ignore_keys: Iterable[str] = tuple(),
):
    """
    Traverses a SMIRNOFF data structure, attempting to convert all
    quantity-defining strings into openff.units.unit.Quantity objects.

    Integers and floats are ignored and not converted into a dimensionless
    ``openff.units.unit.Quantity`` object.

    Parameters
    ----------
    smirnoff_data
        A hierarchical dict structured in compliance with the SMIRNOFF spec
    ignore_keys
        A list of keys to skip when converting strings to quantities

    Returns
    -------
    converted_smirnoff_data
        A hierarchical dict structured in compliance with the SMIRNOFF spec,
        with quantity-defining strings converted to openff.units.unit.Quantity objects
    """
    from pint import DefinitionSyntaxError

    if isinstance(smirnoff_data, dict):
        for key, value in smirnoff_data.items():
            if key in ignore_keys:
                smirnoff_data[key] = value
            else:
                smirnoff_data[key] = convert_all_strings_to_quantity(
                    value,
                    ignore_keys=ignore_keys,
                )
        obj_to_return = smirnoff_data

    elif isinstance(smirnoff_data, list):
        for index, item in enumerate(smirnoff_data):
            smirnoff_data[index] = convert_all_strings_to_quantity(
                item,
                ignore_keys=ignore_keys,
            )
        obj_to_return = smirnoff_data

    elif isinstance(smirnoff_data, int) or isinstance(smirnoff_data, float):
        obj_to_return = smirnoff_data

    else:
        try:
            obj_to_return = object_to_quantity(smirnoff_data)
        except (TypeError, DefinitionSyntaxError):
            obj_to_return = smirnoff_data

    return obj_to_return


@functools.singledispatch
def object_to_quantity(object):
    """
    Attempts to turn the provided object into openmm.unit.Quantity(s).

    Can handle float, int, strings, quantities, or iterators over
    the same. Raises an exception if unable to convert all inputs.

    Parameters
    ----------
    object
        The object to convert to a ``openmm.unit.Quantity`` object.

    Returns
    -------
    converted_object

    """
    # If we can't find a custom type, we treat this as a generic iterator.
    return [object_to_quantity(sub_obj) for sub_obj in object]


@object_to_quantity.register(Quantity)
def _(obj):
    return obj


@object_to_quantity.register(str)
def _(obj):
    import pint

    try:
        return string_to_quantity(obj)
    except pint.errors.UndefinedUnitError:
        raise ValueError


@object_to_quantity.register(int)
@object_to_quantity.register(float)
def _(obj):
    return Quantity(obj)


try:
    import openmm

    from openff.units.openmm import from_openmm

    @object_to_quantity.register(openmm.unit.Quantity)
    def _(obj):
        return from_openmm(obj)

except ImportError:
    pass  # pragma: nocover



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys
from importlib.util import find_spec as find_import_spec

sys.path.insert(0, os.path.abspath("."))


# -- Project information -----------------------------------------------------

project = "OpenFF Units"
copyright = "2022, The Open Force Field Initiative"
author = "The Open Force Field Initiative"


# -- General configuration ---------------------------------------------------


# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autosummary",
    "sphinx.ext.autodoc",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx.ext.extlinks",
    "sphinx.ext.doctest",
    "myst_parser",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "pint": ("https://pint.readthedocs.io/en/stable/", None),
    "openff.toolkit": (
        "https://docs.openforcefield.org/projects/toolkit/en/stable/",
        None,
    ),
    "openff.fragmenter": (
        "https://docs.openforcefield.org/projects/fragmenter/en/stable/",
        None,
    ),
    "openff.interchange": (
        "https://docs.openforcefield.org/projects/interchange/en/stable/",
        None,
    ),
    "openff.qcsubmit": (
        "https://docs.openforcefield.org/projects/qcsubmit/en/stable/",
        None,
    ),
    "openff.evaluator": (
        "https://docs.openforcefield.org/projects/evaluator/en/stable/",
        None,
    ),
    "mdtraj": ("https://www.mdtraj.org/1.9.5/", None),
    "openmm": ("http://docs.openmm.org/latest/api-python/", None),
    "openff.docs": (
        "https://docs.openforcefield.org/en/latest/",
        None,
    ),
}

autosummary_generate = True
autosummary_imported_members = True
autosummary_ignore_module_all = False
autosummary_context = {
    # Modules to exclude from API docs
    "exclude_modules": [
        "openff.units.tests",
        "openff.units.units",
    ]
}

autodoc_default_options = {
    "member-order": "alphabetical",
    "show-inheritance": True,
}
autodoc_preserve_defaults = True
autodoc_inherit_docstrings = False
autodoc_typehints_format = "short"
# Fold the __init__ or __new__ methods into class documentation
autoclass_content = "both"
autodoc_class_signature = "mixed"
# Workaround for autodoc_typehints_format not working for attributes
# see https://github.com/sphinx-doc/sphinx/issues/10290#issuecomment-1079740009
python_use_unqualified_type_names = True

napoleon_google_docstring = True
napoleon_use_param = False
napoleon_use_ivar = True
napoleon_use_rtype = False
napoleon_preprocess_types = True

myst_enable_extensions = [
    "deflist",
    "smartquotes",
    "replacements",
    "dollarmath",
    "colon_fence",
]

# sphinx-notfound-page
# https://github.com/readthedocs/sphinx-notfound-page
# Renders a 404 page with absolute links

if find_import_spec("notfound"):
    extensions.append("notfound.extension")

    notfound_urls_prefix = "/projects/toolkit/en/stable/"
    notfound_context = {
        "title": "404: File Not Found",
        "body": f"""
    <h1>404: File Not Found</h1>
    <p>
        Sorry, we couldn't find that page. This often happens as a result of
        following an outdated link. Please check the
        <a href="{notfound_urls_prefix}">latest stable version</a>
        of the docs, unless you're sure you want an earlier version, and
        try using the search box or the navigation menu on the left.
    </p>
    <p>
    </p>
    """,
    }

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
extensions.append("openff_sphinx_theme")
html_theme = "openff_sphinx_theme"

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
html_sidebars = {
    "**": ["globaltoc.html", "searchbox.html"],
}

# Theme options are theme-specific and customize the look and feel of a
# theme further.
html_theme_options = {
    # Repository integration
    # Set the repo url for the link to appear
    "repo_url": "https://github.com/openforcefield/openff-units",
    # The name of the repo. If must be set if repo_url is set
    "repo_name": "openff-units",
    # Must be one of github, gitlab or bitbucket
    "repo_type": "github",
    # Colour for sidebar captions and other accents. One of
    # openff-blue, openff-toolkit-blue, openff-dataset-yellow,
    # openff-evaluator-orange, aquamarine, lilac, amaranth, grape,
    # violet, pink, pale-green, green, crimson, eggplant, turquoise,
    # or a tuple of three ints in the range [0, 255] corresponding to
    # a position in RGB space.
    "color_accent": "crimson",
    "html_minify": False,
    "html_prettify": False,
    "css_minify": False,
    "globaltoc_depth": 2,
    "globaltoc_include_local": True,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]



import pickle
import random

import pytest
from openff.utilities.testing import skip_if_missing

from openff.units import Quantity, unit


class TestQuantity:
    @skip_if_missing("openmm.unit")
    def test_to_openmm_method(self):
        """
        Test the basic behavior of `Quantity.to_openmm` as an API call. Testing of the underlying
        behavior of the standalone `to_openmm` function is in opeff/units/tests/test_openmm.py.
        """
        from openmm import unit as openmm_unit

        quantity = unit.Quantity(0.5, "nanometer")
        converted = quantity.to_openmm()

        assert converted == openmm_unit.Quantity(0.5, openmm_unit.nanometer)


class TestPickle:
    """Test pickle-based serialization of Quantity, Unit, and Measurement objects

    See:
      * https://github.com/hgrecco/pint/issues/1017
      * https://github.com/openforcefield/openff-evaluator/pull/341

    """

    def test_pickle_unit(self):
        x = unit.kelvin
        y = pickle.loads(pickle.dumps(x))

        assert x == y

    def test_pickle_quantity(self):
        x = 1.0 * unit.kelvin
        y = pickle.loads(pickle.dumps(x))

        assert x == y

    @skip_if_missing("uncertainties")
    def test_pickle_quantity_uncertainties(self):
        x = (1.0 * unit.kelvin).plus_minus(0.05)
        y = pickle.loads(pickle.dumps(x))

        assert x.value == y.value and x.error == y.error


class TestCompChemUnits:
    """Test some non-standard units used in comp chem stacks."""

    @pytest.mark.parametrize(
        "shorthand_string,full_string",
        [
            (
                "2000.0 * kilocalories_per_mole/angstrom**2",
                "2000.0 * kilocalories/mole/angstrom**2",
            ),
            (
                "2000.0 * kilocalorie_per_mole/angstrom**2",
                "2000.0 * kilocalorie/mole/angstrom**2",
            ),
            (
                "2000.0 * kilojoules_per_mole/angstrom**2",
                "2000.0 * kilojoules/mole/angstrom**2",
            ),
            (
                "2000.0 * kilojoule_per_mole/angstrom**2",
                "2000.0 * kilojoule/mole/angstrom**2",
            ),
        ],
    )
    def test_parse_molar_units_string(self, shorthand_string, full_string):
        assert unit.Quantity(shorthand_string) == unit.Quantity(full_string)

    def test_timestep_creation(self):
        # basic sanity check, can I make the unit and does it serialize
        q = 10 * unit.timestep

        assert q.m == 10
        assert str(q) == "10 timestep"

    def test_timestep_compatibility(self):
        # timesteps aren't a form of time
        q = 20 * unit.timestep

        assert not q.is_compatible_with(10 * unit.picosecond)


class TestNMRUnits:
    @pytest.mark.parametrize(
        ["unit_name", "symbol"],
        [
            ("tesla", "T"),
            ("ohm", "Ω"),
            ("henry", "H"),
            ("siemens", "S"),
            ("watt", "W"),
            ("weber", "Wb"),
        ],
    )
    def test_symbols(self, unit_name, symbol):
        number = random.random()
        assert Quantity(f"{number} {unit_name}") == Quantity(f"{number} {symbol}")

    def test_nmr_equivalents(self):
        assert Quantity("1 ohm") == Quantity("1 henry / second")
        assert Quantity("1 ohm") == 1 / Quantity("1 siemens")
        assert Quantity("1 ohm") == Quantity("1 volt / ampere")
        assert Quantity("1 watt") == Quantity("1 joule / second")
        assert Quantity("1 henry") == Quantity("1 weber / ampere")
        assert Quantity("1 tesla") == Quantity("1 weber / meter**2")



"""
Registry of units provided by OpenFF Units.

``unit`` may be used similarly to a module. It makes constants and units of
measure available as attributes. Available units can be found in the
:download:`constants <../../../openff/units/data/constants.txt>` and
:download:`defaults <../../../openff/units/data/defaults.txt>` data files.
"""

import importlib
from importlib.metadata import version
from types import ModuleType
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    # Type checkers can't see lazy-imported objects
    from openff.units.openmm import ensure_quantity
    from openff.units.units import (  # type: ignore[attr-defined]
        DEFAULT_UNIT_REGISTRY,
        Measurement,
        Quantity,
        Unit,
        UnitRegistry,
    )


__version__ = version("acellera-openff-units")

__all__ = [
    "Measurement",
    "Quantity",
    "Unit",
    "ensure_quantity",
    "unit",
]

_objects: dict[str, str] = {
    "ensure_quantity": "openff.units.openmm",
    "Measurement": "openff.units.units",
    "Unit": "openff.units.units",
    "UnitRegistry": "openff.units.units",
    "Quantity": "openff.units.units",
    "DEFAULT_UNIT_REGISTRY": "openff.units.units",
    "unit": "openff.units.units",
}


def __getattr__(name) -> ModuleType:
    """
    Lazily import objects from submodules.

    Taken from openff/toolkit/__init__.py
    """
    module = _objects.get(name)
    if module is not None:
        try:
            return importlib.import_module(module).__dict__[name]
        except ImportError as error:
            raise ImportError from error

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def __dir__():
    """Add _objects to dir()."""
    keys = (*globals().keys(), *_objects.keys())
    return sorted(keys)



__all__ = [
    "MissingOpenMMUnitError",
    "NoneQuantityError",
    "NoneUnitError",
]


class MissingOpenMMUnitError(Exception):
    """Raised when a unit cannot be converted to an equivalent OpenMM unit"""


class NoneQuantityError(Exception):
    """Raised when attempting to convert `None` between unit packages as a quantity object"""


class NoneUnitError(Exception):
    """Raised when attempting to convert `None` between unit packages as a unit object"""



"""
Utility methods for OpenFF Units
"""

from openff.utilities import get_data_file_path

__all__ = [
    "get_defaults_path",
]


def get_defaults_path() -> str:
    """Get the full path to the ``defaults.txt`` file"""
    return get_data_file_path("defaults.txt", "openff.units")



"""
Functions for converting between OpenFF and OpenMM units
"""

import ast
import operator as op
from typing import Literal

from openff.utilities import has_package, requires_package

from openff.units.exceptions import (
    MissingOpenMMUnitError,
    NoneQuantityError,
    NoneUnitError,
)
from openff.units.units import Quantity, Unit

__all__ = [
    "ensure_quantity",
    "from_openmm",
    "openmm_unit_to_string",
    "string_to_openmm_unit",
    "to_openmm",
]

if has_package("openmm.unit"):
    import openmm.unit

    EitherQuantity = Quantity | openmm.unit.Quantity
else:
    EitherQuantity = Quantity  # type: ignore


@requires_package("openmm.unit")
def openmm_unit_to_string(input_unit: "openmm.unit.Unit") -> str:
    """
    Convert a openmm.unit.Unit to a string representation.

    Parameters
    ----------
    input_unit : A openmm.unit
        The unit to serialize

    Returns
    -------
    unit_string : str
        The serialized unit.
    """
    if input_unit is None:
        raise NoneUnitError("Input is None, expected an (OpenMM) Unit object.")

    if input_unit == openmm.unit.dimensionless:
        return "dimensionless"

    if input_unit == openmm.unit.dalton:
        return "g/mol"

    # Decompose output_unit into a tuples of (base_dimension_unit, exponent)
    unit_string = ""

    for unit_component in input_unit.iter_base_or_scaled_units():
        unit_component_name = unit_component[0].name
        # Convert, for example "elementary charge" --> "elementary_charge"
        unit_component_name = unit_component_name.replace(" ", "_")
        if unit_component[1] == 1:
            contribution = f"{unit_component_name}"
        else:
            contribution = f"{unit_component_name}**{int(unit_component[1])}"
        if unit_string == "":
            unit_string = contribution
        else:
            unit_string += f" * {contribution}"

    return unit_string


def _ast_eval(node):
    """
    Performs an algebraic syntax tree evaluation of a unit.

    Parameters
    ----------
    node : An ast parsing tree node

    Raises
    ------
    openff.units.exceptions.MissingOpenMMUnitError
        if the unit is unavailable in OpenMM.
    """

    operators = {
        ast.Add: op.add,
        ast.Sub: op.sub,
        ast.Mult: op.mul,
        ast.Div: op.truediv,
        ast.Pow: op.pow,
        ast.BitXor: op.xor,
        ast.USub: op.neg,
    }

    if isinstance(node, ast.Constant):  # <number>
        return node.value
    elif isinstance(node, ast.BinOp):  # <left> <operator> <right>
        return operators[type(node.op)](_ast_eval(node.left), _ast_eval(node.right))
    elif isinstance(node, ast.UnaryOp):  # <operator> <operand> e.g., -1
        return operators[type(node.op)](_ast_eval(node.operand))
    elif isinstance(node, ast.Name):
        # see if this is a openmm unit
        try:
            b = getattr(openmm.unit, node.id)
        except AttributeError:
            raise MissingOpenMMUnitError(node.id)
        return b
    # TODO: This toolkit code that had a hack to cover some edge behavior
    #       not clear which tests trigger it
    elif isinstance(node, ast.List):
        return ast.literal_eval(node)
    else:
        raise TypeError(node)


def string_to_openmm_unit(unit_string: str) -> "openmm.unit.Unit":
    """
    Deserializes a openmm.unit.Quantity from a string representation, for
    example: "kilocalories_per_mole / angstrom ** 2"


    Parameters
    ----------
    unit_string : dict
        Serialized representation of a openmm.unit.Quantity.

    Returns
    -------
    output_unit: openmm.unit.Quantity
        The deserialized unit from the string

    Raises
    ------
    openff.units.exceptions.MissingOpenMMUnitError
        if the unit is unavailable in OpenMM.
    """
    if unit_string == "standard_atmosphere":
        return openmm.unit.atmosphere

    output_unit = _ast_eval(ast.parse(unit_string, mode="eval").body)
    return output_unit


@requires_package("openmm.unit")
def from_openmm(openmm_quantity: "openmm.unit.Quantity") -> Quantity:
    """Convert an OpenMM ``Quantity`` to an OpenFF ``Quantity``

    :class:`openmm.unit.quantity.Quantity` from OpenMM and
    :class:`openff.units.Quantity` from this package both represent a numerical
    value with units.

    Examples
    --------

    >>> from openff.units import Quantity as OpenFFQuantity
    >>> from openff.units.openmm import from_openmm
    >>> from openmm import unit
    >>> length = unit.Quantity(9.0, unit.angstrom)
    >>> from_openmm(length)
    <Quantity(9.0, 'angstrom')>
    >>> assert isinstance(from_openmm(length), OpenFFQuantity)

    """
    if openmm_quantity is None:
        raise NoneQuantityError("Input is None, expected an (OpenMM) Quantity object.")

    if isinstance(openmm_quantity, list):
        openmm_quantity = openmm.unit.Quantity(openmm_quantity)
    openmm_unit_ = openmm_quantity.unit
    openmm_value = openmm_quantity.value_in_unit(openmm_unit_)

    target_unit = openmm_unit_to_string(openmm_unit_)
    target_unit = Unit(target_unit)

    return Quantity(openmm_value, target_unit)


@requires_package("openmm.unit")
def to_openmm(quantity: Quantity) -> "openmm.unit.Quantity":
    """Convert an OpenFF ``Quantity`` to an OpenMM ``Quantity``

    :class:`openmm.unit.quantity.Quantity` from OpenMM and
    :class:`openff.units.Quantity` from this package both represent a numerical
    value with units. The units available in the two packages differ; when a
    unit is missing from the target package, the resulting quantity will be in
    base units (kg/m/s/A/K/mole), which are shared between both packages. This
    may cause the resulting value to be slightly different to the input due to
    the limited precision of floating point numbers.

    Examples
    --------

    >>> from openff.units import unit
    >>> from openff.units.openmm import to_openmm
    >>> from openmm import unit as openmm_unit
    >>> length = unit.Quantity(9.0, unit.angstrom)
    >>> to_openmm(length)
    9.0 A
    >>> assert isinstance(to_openmm(length), openmm_unit.Quantity)

    """
    if quantity is None:
        raise NoneQuantityError("Input is None, expected an (OpenFF) Quantity object.")

    def to_openmm_inner(quantity) -> "openmm.unit.Quantity":
        value = quantity.m

        unit_string = str(quantity.units._units)
        openmm_unit_ = string_to_openmm_unit(unit_string)

        return value * openmm_unit_

    assert isinstance(quantity, Quantity)

    try:
        return to_openmm_inner(quantity)
    except MissingOpenMMUnitError:
        return to_openmm_inner(quantity.to_base_units())


@requires_package("openmm.unit")
def _ensure_openmm_quantity(
    unknown_quantity: EitherQuantity,
) -> "openmm.unit.Quantity":
    if "openmm" in str(type(unknown_quantity)):
        from openmm import unit as openmm_unit

        if isinstance(unknown_quantity, openmm_unit.Quantity):
            return unknown_quantity
        else:
            raise ValueError(f"Failed to process input of type {type(unknown_quantity)}.")
    elif isinstance(unknown_quantity, Quantity):
        return to_openmm(unknown_quantity)
    else:
        import openmm.unit

        try:
            return openmm.unit.Quantity(
                unknown_quantity,
                openmm.unit.dimensionless,
            )
        except Exception as e:
            raise ValueError(f"Failed to process input of type {type(unknown_quantity)}.") from e


def _ensure_openff_quantity(
    unknown_quantity: EitherQuantity,
) -> Quantity:
    if isinstance(unknown_quantity, Quantity):
        return unknown_quantity
    elif "openmm" in str(type(unknown_quantity)):
        import openmm.unit

        if isinstance(unknown_quantity, openmm.unit.Quantity):
            return from_openmm(unknown_quantity)
        else:
            raise ValueError(f"Failed to process input of type {type(unknown_quantity)}.")
    else:
        try:
            return Quantity(
                unknown_quantity,
                "dimensionless",
            )
        except Exception as e:
            raise ValueError(f"Failed to process input of type {type(unknown_quantity)}.") from e


def ensure_quantity(
    unknown_quantity: EitherQuantity,
    type_to_ensure: Literal["openmm", "openff"],
) -> EitherQuantity:
    """
    Given a quantity that could be of a variety of types, attempt to coerce into a given type.

    Examples
    --------

    >>> import numpy
    >>> from openmm import unit as openmm_unit
    >>> from openff.units import unit
    >>> from openff.units.openmm import ensure_quantity
    >>> # Create a 9 Angstrom quantity with each registry
    >>> length1 = unit.Quantity(9.0, unit.angstrom)
    >>> length2 = openmm_unit.Quantity(9.0, openmm_unit.angstrom)
    >>> # Similar quantities are be coerced into requested type
    >>> assert type(ensure_quantity(length1, "openmm")) == openmm_unit.Quantity
    >>> assert type(ensure_quantity(length2, "openff")) == unit.Quantity
    >>> # Seemingly-redundant "conversions" short-circuit
    >>> assert ensure_quantity(length1, "openff") == ensure_quantity(length2, "openff")
    >>> assert ensure_quantity(length1, "openmm") == ensure_quantity(length2, "openmm")
    >>> # NumPy arrays and some primitives are automatically up-converted to `Quantity` objects
    >>> # Note that their units are set to "dimensionless"
    >>> ensure_quantity(numpy.array([1, 2]), "openff")
    <Quantity([1 2], 'dimensionless')>
    >>> ensure_quantity(4.0, "openmm")
    4.0 dimensionless

    """
    if type_to_ensure == "openmm":
        return _ensure_openmm_quantity(unknown_quantity)
    elif type_to_ensure == "openff":
        return _ensure_openff_quantity(unknown_quantity)
    else:
        raise ValueError(
            f"Unsupported `type_to_ensure` found. Given {type_to_ensure}, expected 'openff' or "
            "'openmm'."
        )



import pytest
from dummy import object_to_quantity

from openff.units import Quantity, Unit


def test_function_can_be_defined():
    """
    Just make sure a function can be defined using these classes.

    Safeguard against i.e. https://github.com/openforcefield/openff-toolkit/issues/1632

    Type checkers might not be happy.
    """

    def dummy_function(
        unit: Unit,
        quantity: Quantity,
        extra: Unit | Quantity | str | None = None,
    ):
        return f"{unit} {quantity} {extra}"


@pytest.mark.parametrize(
    "input, output",
    [
        ("1.0 * kilocalories_per_mole", Quantity(1.0, "kilocalories_per_mole")),
        (Quantity("2.0 * angstrom"), Quantity(2.0, "angstrom")),
        (3.0 * Unit("nanometer"), Quantity(3.0, "nanometer")),
        (4, Quantity(4.0)),
        (5.0, Quantity(5.0)),
    ],
)
def test_object_to_quantity(input, output):
    assert object_to_quantity(input) == output



"""
Core classes for OpenFF Units
"""

import uuid
import warnings
from typing import TYPE_CHECKING

import pint
from openff.utilities import requires_package
from pint import Measurement as _Measurement
from pint import Quantity as _Quantity
from pint import Unit as _Unit

from openff.units.utilities import get_defaults_path

if TYPE_CHECKING:
    import openmm.unit

__all__ = (
    "DEFAULT_UNIT_REGISTRY",
    "Measurement",
    "Quantity",
    "Unit",
    "unit",
)


class Unit(pint.UnitRegistry.Unit):
    """A unit of measure."""

    pass


class Quantity(pint.UnitRegistry.Quantity):
    """A value with associated units."""

    def __dask_tokenize__(self):
        return uuid.uuid4().hex

    @staticmethod
    def _dask_finalize(results, func, args, units):
        values = func(results, *args)
        return Quantity(values, units)


@requires_package("openmm")
def _to_openmm(self) -> "openmm.unit.Quantity":
    """Convert the quantity to an ``openmm.unit.Quantity``.

    Returns
    -------
    openmm_quantity : openmm.unit.quantity.Quantity
        The OpenMM compatible quantity.
    """
    from openff.units.openmm import to_openmm

    return to_openmm(self)


class Measurement(pint.UnitRegistry.Measurement):
    """A value with associated units and uncertainty."""

    def __dask_tokenize__(self):
        return uuid.uuid4().hex

    @staticmethod
    def _dask_finalize(results, func, args, units):
        values = func(results, *args)
        return Measurement(values, units)


class UnitRegistry(pint.UnitRegistry):
    _quantity_class = Quantity
    _unit_class = Unit
    _measurement_class = Measurement


DEFAULT_UNIT_REGISTRY = UnitRegistry(get_defaults_path())

unit = DEFAULT_UNIT_REGISTRY

Unit: type[_Unit] = DEFAULT_UNIT_REGISTRY.Unit
Quantity: type[_Quantity] = DEFAULT_UNIT_REGISTRY.Quantity
Measurement: type[_Measurement] = DEFAULT_UNIT_REGISTRY.Measurement

pint.set_application_registry(DEFAULT_UNIT_REGISTRY)

Quantity.to_openmm = _to_openmm  # type: ignore[attr-defined]

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    Quantity([])



import pytest
from openff.utilities.testing import skip_if_missing
from openff.utilities.utilities import has_package

from openff.units import Quantity, unit
from openff.units.exceptions import NoneQuantityError, NoneUnitError
from openff.units.openmm import ensure_quantity, from_openmm

if has_package("openmm.unit"):
    from openmm import unit as openmm_unit

    from openff.units.openmm import (
        openmm_unit_to_string,
        string_to_openmm_unit,
        to_openmm,
    )

    openmm_quantitites = [
        4.0 * openmm_unit.nanometer,
        5.0 * openmm_unit.angstrom,
        1.0 * openmm_unit.elementary_charge,
        0.5 * openmm_unit.erg,
        1.0 * openmm_unit.dimensionless,
        0.5 * openmm_unit.dalton,
    ]

    pint_quantities = [
        4.0 * unit.nanometer,
        5.0 * unit.angstrom,
        1.0 * unit.elementary_charge,
        0.5 * unit.erg,
        1.0 * unit.dimensionless,
        0.5 * unit.gram / unit.mol,
    ]
else:
    # Must be defined as something, despite not being used, because pytest
    # inspect the contents of pytest.mark.parametrize during collection;
    # otherwise NameErrors will be raised. Finding a way to skip _collection_
    # would be more elegant than mocking a module
    class openmm_unit:  # type: ignore[no-redef]
        kilojoule = 1
        kilojoule_per_mole = 1
        kilocalories_per_mole = 1
        angstrom = 1
        nanometer = 1
        meter = 1
        picosecond = 1
        joule = 1
        mole = 1
        dimensionless = 1
        second = 1
        kelvin = 1
        kilogram = 1
        ampere = 1

    openmm_quantitites = []
    pint_quantities = []


@skip_if_missing("openmm.unit")
class TestOpenMMUnits:
    @pytest.mark.parametrize(
        "openmm_quantity,pint_quantity",
        [(s, p) for s, p in zip(openmm_quantitites, pint_quantities)],
    )
    def test_openmm_to_pint(self, openmm_quantity, pint_quantity):
        """Test conversion from OpenMM Quantity to pint Quantity."""
        converted_pint_quantity = from_openmm(openmm_quantity)

        assert pint_quantity == converted_pint_quantity

    def test_from_openmm_quantity_none(self):
        with pytest.raises(NoneQuantityError, match=r"Input is None.*OpenMM.*Quantity"):
            from_openmm(None)

    def test_to_openmm_quantity_none(self):
        with pytest.raises(NoneQuantityError, match=r"Input is None.*OpenFF.*Quantity"):
            to_openmm(None)

    def test_openmm_unit_to_string_none(self):
        with pytest.raises(NoneUnitError, match=r"Input is None.*OpenMM.*Unit"):
            openmm_unit_to_string(None)

    @pytest.mark.parametrize(
        "openmm_quantity,pint_quantity",
        [(s, p) for s, p in zip(openmm_quantitites, pint_quantities)],
    )
    def test_pint_to_openmm(self, openmm_quantity, pint_quantity):
        """Test conversion from pint Quantity to OpenMM Quantity."""
        converted_openmm_quantity = to_openmm(pint_quantity)

        assert openmm_quantity == converted_openmm_quantity

    @skip_if_missing("openmm.unit")
    @pytest.mark.parametrize(
        "openmm_unit_,unit_str",
        [
            (openmm_unit.kilojoule_per_mole, "mole**-1 * kilojoule"),
            (
                openmm_unit.kilocalories_per_mole / openmm_unit.angstrom**2,
                "angstrom**-2 * mole**-1 * kilocalorie",
            ),
            (
                openmm_unit.joule / (openmm_unit.mole * openmm_unit.nanometer**2),
                "nanometer**-2 * mole**-1 * joule",
            ),
            (
                openmm_unit.picosecond ** (-1),
                "picosecond**-1",
            ),
            (openmm_unit.dimensionless, "dimensionless"),
            (openmm_unit.second, "second"),
            (openmm_unit.angstrom, "angstrom"),
        ],
    )
    def test_openmm_unit_string_roundtrip(self, openmm_unit_, unit_str):
        assert openmm_unit_to_string(openmm_unit_) == unit_str

        assert unit_str == openmm_unit_to_string(string_to_openmm_unit(unit_str))

    @pytest.mark.parametrize(
        "openff_quantity,openmm_quantity",
        [
            (300.0 * unit.kelvin, 300.0 * openmm_unit.kelvin),
            (
                1.5 * unit.kilojoule,
                1.5 * openmm_unit.kilojoule,
            ),
            (1.0 / unit.meter, 1.0 / openmm_unit.meter),
        ],
    )
    def test_openmm_roundtrip(self, openff_quantity, openmm_quantity):
        assert openmm_quantity == to_openmm(openff_quantity)
        assert openff_quantity == from_openmm(openmm_quantity)

        assert openff_quantity == from_openmm(to_openmm(openff_quantity))

    @pytest.mark.parametrize(
        "from_openff_quantity,to_openmm_quantity",
        [
            (
                1.0 * unit.k_B,
                1.380649e-23
                * openmm_unit.meter**2
                * openmm_unit.kilogram
                * openmm_unit.second**-2
                * openmm_unit.kelvin**-1,
            ),
            (
                1.0 / unit.pi,
                0.31830988618 * openmm_unit.dimensionless,
            ),
            (
                1.0 * unit.avogadro_constant,
                6.02214076e23 / openmm_unit.mole,
            ),
            (
                1.0 * unit.vacuum_permittivity,
                8.85418782e-12
                * openmm_unit.meter**-3
                * openmm_unit.kilogram**-1
                * openmm_unit.second**4
                * openmm_unit.ampere**2,
            ),
        ],
    )
    def test_openmm_unit_constants(self, from_openff_quantity, to_openmm_quantity):
        """Test conversion of units that do not exist in the OpenMM registry

        For these units, the magnitude/value may change due to limited floating
        point precision. We therefore do not require the final values to be
        exact - we accept a variance of one part in one million."""
        converted = to_openmm(from_openff_quantity)

        assert abs(converted - to_openmm_quantity) < (
            # Multiply by dimensionless to ensure result is an openmm quantity
            1.0e-6 * to_openmm_quantity * openmm_unit.dimensionless
        )


@skip_if_missing("openmm.unit")
class TestEnsureType:
    from openff.units import unit

    @pytest.mark.parametrize(
        "registry",
        ["openmm", "openff"],
    )
    def test_ensure_units(self, registry):
        x = unit.Quantity(4.0, unit.angstrom)
        y = openmm_unit.Quantity(4.0, openmm_unit.angstrom)

        assert ensure_quantity(x, registry) == ensure_quantity(y, registry)

    def test_unsupported_type(self):
        x = unit.Quantity(4.0, unit.angstrom)

        with pytest.raises(ValueError, match=r"Unsupported.*type_to_ensure.*pint"):
            ensure_quantity(x, "pint")

    def test_short_circuit(self):
        x = unit.Quantity(4.0, unit.angstrom)
        y = openmm_unit.Quantity(4.0, openmm_unit.angstrom)

        assert id(ensure_quantity(x, "openff")) == id(x)
        assert id(ensure_quantity(y, "openmm")) == id(y)

    @pytest.mark.parametrize(
        "value",
        [
            1,
            2.0,
        ],
    )
    def test_primitives(self, value):
        assert ensure_quantity(value, "openff") == unit.Quantity(value, unit.dimensionless)
        assert ensure_quantity(value, "openmm") == openmm_unit.Quantity(
            value, openmm_unit.dimensionless
        )

    @pytest.mark.parametrize("use_numpy", [True, False])
    def test_array(self, use_numpy):
        import numpy

        value = [4, 5]

        if use_numpy:
            value = numpy.asarray(value)

        numpy.testing.assert_allclose(
            ensure_quantity(value, "openff").m,
            Quantity(value, unit.dimensionless).m,
        )

        numpy.testing.assert_equal(
            numpy.array(ensure_quantity(value, "openmm")),
            numpy.array(openmm_unit.Quantity(value, openmm_unit.dimensionless)),
        )



"""
A test package used to ensure that downstream packages can do basic things with openff-units.
"""

from setuptools import setup

setup(
    version="0.0.0",
    name="openff-units-downstream-dummy",
)



import numpy

class Unit:
    def __init__(self, *args, **kwargs): ...
    def __mul__(self, other: int | float | numpy.ndarray) -> Quantity: ...
    __rmul__ = __mul__

class Quantity:
    def __init__(self, *args, **kwargs): ...
    def to(self, unit: str | Unit = "dimensionless") -> Quantity: ...
    def to_base_units(self, unit: str | Unit = "dimensionless") -> Quantity: ...
    def is_compatible_with(self, unit: str | Unit) -> bool: ...
    @property
    def u(self) -> Unit: ...

    # lists are cast to ndarray
    @property
    def m(self) -> float | int | numpy.ndarray: ...
    def m_as(self, unit: str | Unit) -> float | int | numpy.ndarray: ...

    magnitude = m
    units = u



from collections.abc import Generator

import numpy
from openmm.unit.baseunit import BaseUnit

class Quantity:
    def __init__(self, *args, **kwargs) -> None: ...
    @property
    def unit(self) -> Unit: ...
    def value_in_unit(self, unit: Unit | str) -> float | int | list | numpy.ndarray: ...

class Unit:
    def __init__(self, name: str) -> None: ...
    def iter_base_or_scaled_units(
        self,
    ) -> Generator[tuple[BaseUnit, int], None, None]: ...
    def __mul__(self, other: float | int | list | numpy.ndarray) -> Quantity: ...
    def __rmul__(self, other: float | int | list | numpy.ndarray) -> Quantity: ...

dalton = Unit("dalton")
dimensionless = Unit("dimensionless")
atmosphere = Unit("atmosphere")



import random

import pytest
from openff.utilities.testing import skip_if_missing

from openff.units.elements import MASSES, NUMBERS, SYMBOLS


@skip_if_missing("openmm.unit")
@pytest.mark.parametrize("atomic_number", [*range(1, 100)])
def test_openmm_parity(atomic_number):
    from openmm.app.element import Element

    openmm_element = Element.getByAtomicNumber(atomic_number)

    assert MASSES[atomic_number].m == openmm_element.mass._value
    assert SYMBOLS[atomic_number] == openmm_element.symbol


def test_basic():
    assert MASSES[1].m == pytest.approx(1.007947)
    assert MASSES[6].m == pytest.approx(12.01078)

    assert SYMBOLS[1] == "H"
    assert SYMBOLS[6] == "C"

    assert NUMBERS["Cl"] == 17
    assert NUMBERS["Cf"] == 98


def test_symbol_roundtrip():
    number = random.randrange(1, 100)

    assert NUMBERS[SYMBOLS[number]] == number
